#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <math.h>
#include <time.h>
#include <cassert>
#include <cv.h>
#include <highgui.h>
#include "BlobResult.h"

using namespace std;

// Function headers
bool alreadyHaveIt(int j,int imageIndices[12]);
void RGBtoHSV(double r, double g, double b, double *h, double *s, double *v);

void processSet(IplImage* frame, IplImage* f_p , int index);
bool isRed(double h, double s, double v, double r, double g, double b);
bool isGreen(double h, double s, double v, double r, double g, double b);
bool isPurple(double h, double s, double v, double r, double g, double b);

void setGlobals();  // the code appears at the very bottom of this file

char wndname[] = "Single card image";
char wndname2[] = "Painted image";

char *setwnd[] = {"set0","set1","set2","set3","set4","set5","set6","set7","set8","set9","set10","set11"};

// here, static refers to the fact that these variables
// are local to this file (they're not available in other files,
// which is safer because it allows reuse of their names)
static CvPoint old_click_pt;
static CvPoint new_click_pt;

// global points for drawing boxes, defining loops, etc.
static CvPoint pt4;
static CvPoint pt5;

// info on each of the set cards
enum Color {RED, GREEN, PURPLE};
enum Texture {FILLED, EMPTY, STRIPED};
enum Shape {OVAL, SQUIGGLE, DIAMOND};
static int number;
static Color color;
static Texture texture;
static Shape shape;
static int numbers[13];
static Color colors[13];
static Texture textures[13];
static Shape shapes[13];

// more global variables
// these are simply recording all the different states you
// can put the program in via the keyboard
static bool getImageFromFile = false;
static bool getImageAgainFromFile = false;
static bool getImage0FromVideoBMPs = false;
static bool saveNextImage = false;
static bool showing_savedImage = false;
static bool saveOnlyOneImageToFile = true;
static bool firstTimeForSet = true;

static const char* generalFilename = "../VideoBMPs/image%05d.bmp";
static char* filename = new char[150];
static char* filename2 = new char[150];
static int frameNumber = 0;
static int keyCode = -1;

// pointers to images in memory
static IplImage* frame;
static IplImage* frame_painted;

static IplImage** setimages = NULL;
static IplImage** setimages_painted = NULL;

// handle mouse clicks here
void mouse_callback(int event, int x, int y, int flags, void* param)
{
    if (event == CV_EVENT_LBUTTONDOWN)
    {
        // cout << "(x,y) = (" << x << "," << y << ")" << endl;

        // reset old_click_pt
        old_click_pt.x = new_click_pt.x;
        old_click_pt.y = new_click_pt.y;

        // get new click point -- note the coordinate change in y
        new_click_pt.x = x;  // coming in from the window system
        new_click_pt.y = frame->height-y;
        // window system and images have different y axes
    }
}

// print pixel information
void pixelInformation(IplImage* f)
{
    if (f == NULL) return;

    // see if things have changed...
    if (new_click_pt.x != old_click_pt.x || new_click_pt.y != old_click_pt.y)
    // has it changed?
    {
        if (new_click_pt.x < 0 || new_click_pt.x > 319 ||
                new_click_pt.y < 0 || new_click_pt.y > 239) return;
        // draw a line from old to new
        // set the old = the new
        old_click_pt.x = new_click_pt.x;
        old_click_pt.y = new_click_pt.y;
        // print the RGB values and coordinates of the newly clicked point:
        // note that blue is #0, green is #1, and red is #2 !!

        /* note that this uses the undistorted frame right here... */

        // see the processSet function for an explanation
        // of why we need "realy"
        int realy = new_click_pt.y;
        //if (getImageAgainFromFile || getImageFromFile) 
        {
            realy = f->height - realy;
        }
        uchar* pixel = cvPtr2D(f,realy,new_click_pt.x);

        cout << "(x,y) = (" << new_click_pt.x << "," << new_click_pt.y << ") with (r,g,b) = ("
             << (int)(pixel[2]) << "," << (int)(pixel[1]) << "," << (int)(pixel[0]) << ")";

        double hue,sat,val;
        RGBtoHSV((double)pixel[2],(double)pixel[1],(double)pixel[0],&hue,&sat,&val);

        cout << " and (h,s,v) = (" << hue << "," << sat << "," << val << ")" << endl;
    }
}


int main(int argc, char** argv)
{
    // in case you want to log things...
    //ofstream logfile("logfile_out.txt");
    //ifstream loginfile("logfile_in.txt");

    // set up the globals...
    setGlobals();

    char szMsg[80];

    /* 
     * set up the video windows
     */


    for (int i=0 ; i<12 ; ++i)
        cvNamedWindow(setwnd[i], CV_WINDOW_AUTOSIZE);

    cvNamedWindow(wndname2, CV_WINDOW_AUTOSIZE);
    cvNamedWindow(wndname, CV_WINDOW_AUTOSIZE);

    /*
     * set up the mouse input
     */
    cvSetMouseCallback(wndname, mouse_callback);

    // main loop
    while (1)
    {
        // click only on undis_image
        pixelInformation(frame);//_painted);

        //handle key presses
        /*
         * quits
         */
        if((keyCode = cvWaitKey(10)) == ((int)'q')) {          
            break;                       // quit everything
        }
        /*
         * plays the game of Set
         */
        else if (keyCode == (int)'S') {                           
            getImageFromFile = false;
            char filenameset[500]; // for filename...

            /* is this the first play? -- we need to open 12 windows... */
            if (firstTimeForSet == true)
            {
                srand((unsigned)time(NULL));

                cout << "Setting up set!" << endl;
                setimages = new IplImage*[12]; // should delete this somewhere...
                setimages_painted = new IplImage*[12];
                for (int i=0 ; i<12 ; ++i)
                {
                    sprintf(filenameset,generalFilename,i);
                    //cvReleaseImage(&frame);
                    setimages[i] = cvLoadImage(filenameset); // load image0 12 times.
                    setimages_painted[i] = cvLoadImage(filenameset);
                    cout << "Setting up window " << i << " named " << setwnd[i] << endl;
                    cvShowImage(setwnd[i], setimages[i]);
                }   

                firstTimeForSet = false;
            }
            else {
                cout << "Reading in random images..." << endl;
                int imageIndices[12];
                for (int i=0 ; i<12 ; ++i) { imageIndices[i] = -1; }

                for (int i=0 ; i<12 ; ++i)
                {
                    while (1) {
                        int randImage = (int)(81.0*rand()/RAND_MAX);
                        if (randImage < 0 || randImage > 80) continue;
                        cout << "  randImage is " << randImage << endl;
                        bool b = alreadyHaveIt(randImage,imageIndices);
                        if (!b)
                           imageIndices[i] = randImage;
                        else
                            continue;
                        break;
                    }
                    sprintf(filenameset,generalFilename,imageIndices[i]);
                    cvReleaseImage(&(setimages[i]));
                    setimages[i] = cvLoadImage(filenameset);
                    cvReleaseImage(&(setimages_painted[i]));
                    setimages_painted[i] = cvLoadImage(filenameset);
                    cout << "Setting up window " << i << " named " <<
                        setwnd[i] << " with image #" << imageIndices[i] << endl;
                    cvShowImage(setwnd[i], setimages[i]);
                }
            }
            //Find Sets
            for(int i=0; i<12; ++i)
            {
                processSet(setimages[i],setimages_painted[i], i);
            }
            bool found = false;
            for (int i = 0; i < 12; i++)
                for (int j = i+1; j < 12; j++)
                    for (int k = j+1; k < 12; k++)
                    {
                        if  (((numbers[i] == numbers[j]) +
                              (numbers[j] == numbers[k]) +
                              (numbers[k] == numbers[i])) % 3 == 0)
                        if   (((colors[i] == colors[j]) +
                               (colors[j] == colors[k]) +
                               (colors[k] == colors[i])) % 3 == 0)
                        if (((textures[i] == textures[j]) +
                             (textures[j] == textures[k]) +
                             (textures[k] == textures[i])) % 3 == 0)
                        if   (((shapes[i] == shapes[j]) +
                               (shapes[j] == shapes[k]) +
                               (shapes[k] == shapes[i])) % 3 == 0)

                        {
                            cout << i << ", " << j << ", " << k <<
                                " is a set." << endl;
                            if (!found)
                            {
                                found = true;
                                CvPoint p1, p2;
                                p1.x = 1;
                                p1.y = 1;
                                p2.x = setimages_painted[i]->width - 2;
                                p2.y = setimages_painted[i]->height - 2;
                                cvRectangle(setimages[i], p1, p2,
                                        CV_RGB(0xff,0x00,0x00), 2);
                                cvRectangle(setimages[j], p1, p2,
                                        CV_RGB(0xff,0x00,0x00), 2);
                                cvRectangle(setimages[k], p1, p2,
                                        CV_RGB(0xff,0x00,0x00), 2);
                            }
                        }
                    }
            for(int i=0; i<12; ++i)
            {
                cvShowImage(setwnd[i], setimages[i]);
            }
        } 
        /*
         * the spacebar advances the image in the VideoBMPs directory
         */
        else if (keyCode == (int)' ') {
            int i;
            if (showing_savedImage != true)
            {
                showing_savedImage = true;
                frame == NULL;
            }
            else
            {
                sscanf(filename2,generalFilename,&i);
                sprintf(filename2,generalFilename,++i);
                cvReleaseImage(&frame);
                frame = cvLoadImage(filename2);
                cvReleaseImage(&frame_painted);
                frame_painted = cvLoadImage(filename2);
            }
            if (frame == NULL) 
            {
                i = 0;
                sprintf(filename2,generalFilename,i);
                cvReleaseImage(&frame);
                frame = cvLoadImage(filename2);
                cvReleaseImage(&frame_painted);
                frame_painted = cvLoadImage(filename2);
            }
            cout << "Single image: " << i << endl;
            processSet(frame, frame_painted, 12);
            cvShowImage(wndname, frame);
            cvShowImage(wndname2, frame_painted);
        }
    }

    // release things before quitting
    cvDestroyWindow("result");
    delete[] setimages;
    delete[] setimages_painted;

} // end of main

bool alreadyHaveIt(int j,int imageIndices[12])
{
    for (int i=0 ; i<12 ; ++i)
    {
        if (imageIndices[i] == j) return true;
    }
    return false;
}

// just in case you'd like to use these...
inline double mymin(double a, double b) { return a<b?a:b; }

inline double mymax(double a, double b) { return a>b?a:b; }

// read about HSV at http://en.wikipedia.org/wiki/HSV_color_space
// and http://www.cs.rit.edu/~ncs/color/t_convert.html#RGB%20to%20HSV%20&%20HSV%20to%20RGB
//
// feel free to grab code from above or elsewhere online, if you'd like
// be sure you understand what RANGE _your_ code for HSV is "returning"
// in the h, s, and v pointers...
void RGBtoHSV(double r, double g, double b, double *h, double *s, double *v)
{
    double maxRGB;
    double minRGB;
    double delta;

    // get max and min
    if(r >= g && r >= b)
    {
        maxRGB = r;
        if(g >= b) minRGB = b;
        else minRGB = g;
    }
    else if(g >= b)
    {
        maxRGB = g;
        if(r >= b) minRGB = b;
        else minRGB = r;
    }
    else
    {
        maxRGB = b;
        if(r >= g) minRGB = g;
        else minRGB = r;
    }

    delta = maxRGB - minRGB;
    *v = maxRGB/255.0;

    if (maxRGB <= 0.0)
    {
        *h = -1;
        *s = 0;
        return;
    }

    *s = delta/maxRGB;   // better scaling than 255!
    if(delta == 0)
    {
        *h = -1;
        return;
    }

/*    if(r == maxRGB)
        *h = (g - b)/delta;
    else if (g == maxRGB)
        *h = (1.0/3.0) + (b-r)/delta;
    else
        *h = (2.0/3.0) + (r-g)/delta;
 */
    if(r == maxRGB)
        *h = (g - b)/delta;
    else if (g == maxRGB)
        *h = 2 + (b-r)/delta;
    else
        *h = 4 + (r-g)/delta;
    *h = *h / 6;
    if (*h < 0) *h += 1;
}

// here is where we seek to label the particular set card and indicate what
// we've found the input named "frame" is the source location of our pixels
// the "output" named f_p is the destination for drawing
void processSet(IplImage* frame, IplImage* f_p, int index)
{
    // some useful local variables...
    int r,g,b;
    double hue,sat,val;

    // pixels are handled as uchar*'s
    uchar* currentPixel;     // a pixel from the input, "frame"
    uchar* painted;   // a pixel from the output, "painted"

    // this draws a rectangle from pt4 to pt5 with the color
    // specified by the three RGB values in the CV_RGB macro
    // the R, G, and B channels are all 0 to 255...

    // see the caution below about inverting the y axis in order to
    // use graphics on the live video stream...
    cvRectangle(f_p, pt4, pt5, CV_RGB(250,250,0), 1); // 1 == thickness
    //cvLine(f_p, threeTop, threeBot, CV_RGB(250, 250, 250), 1); // line
    //cvCircle(f_p, pt4, 10,  CV_RGB(255,255,0), /*thickness*/ 1);

    // This loops through the pixels in the rectangle specified
    // by pt4 and pt5 (see setGlobals for their corners)

    int xlow = (int)mymin(pt4.x,pt5.x);
    int xhi = (int)mymax(pt4.x,pt5.x);
    int ylow = (int)mymin(pt4.y,pt5.y);
    int yhi = (int)mymax(pt4.y,pt5.y);

    //int matrix[300][250];
    //assert(xhi - xlow < 300 && yhi - ylow < 250);
    IplImage* mask = cvCreateImage(cvSize(frame->width,
            frame->height), IPL_DEPTH_8U, 1);

    int y, realy;

    // The # of pixels that are classified as red, green, or blue/purple.
    int reds = 0;
    int blues = 0;
    int greens = 0;

    for (int x = 0; x<mask->width; x++)
        for (int y = 0; y<mask->height; y++)
            ((uchar*)(mask->imageData + mask->widthStep*(y)))[x] = 0;
    for (int x = xlow; x<xhi; x++)      //the x direction goes left to right
    {
        for (int y = ylow; y<yhi; y++)  //the y direction goes top to bottom
        {
            // Unfortunately, access to the pixels is different depending on
            // the source of the pixels: static images and video reverse their
            // y orientations.  We handle this here...
            realy = y;
            if (getImageAgainFromFile || getImageFromFile) 
            {
                //realy = frame->height - realy;  // reverse if static image
            }

            // note that we need to use "realy" in each of these cases
            currentPixel = cvPtr2D(frame,realy,x); // the pixel we're looking @
            
            painted = cvPtr2D(f_p,realy,x);    // the pixel we (might) change

            r = (int)(currentPixel[2]);  // the red comp. of the pixel
            g = (int)(currentPixel[1]);  // the green comp. of the pixel
            b = (int)(currentPixel[0]);  // the blue comp. of the pixel

            // you might want to call this here!
            RGBtoHSV(r,g,b,&hue,&sat,&val);

            if(isRed(hue,sat,val,r,g,b))
            {
                reds++;
                painted[0] = 0;  // sets blue to 0
                painted[1] = 0;    // sets green to 0
                painted[2] = 255;    // sets red to 255
                ((uchar*)(mask->imageData + mask->widthStep*(y)))[x] = 255;
                //matrix[y-ylow][x-xlow]=1;
            }
            else if(isGreen(hue,sat,val,r,g,b))
            {
                greens++;
                painted[0] = 0;  // sets blue to 0
                painted[1] = 255;    // sets green to 255
                painted[2] = 0;    // sets red to 0
                ((uchar*)(mask->imageData + mask->widthStep*(y)))[x] = 255;
                //matrix[y-ylow][x-xlow]=1;
            }
            else if(isPurple(hue,sat,val,r,g,b))
            {
                blues++;
                painted[0] = 255;  // sets blue to 255
                painted[1] = 0;    // sets green to 0
                painted[2] = 0;    // sets red to 0
                ((uchar*)(mask->imageData + mask->widthStep*(y)))
                        [x] = 255;
                //matrix[y-ylow][x-xlow]=1;
            }
            else if(((uchar*)(mask->imageData + mask->widthStep*(y-1)))[x] != 0)
                ((uchar*)(mask->imageData + mask->widthStep*(y)))[x] = 254;
        }
    }
    CvScalar maxColor;
    if (reds > blues && reds > greens)
    {
        colors[index] = RED;
        maxColor = CV_RGB(0xff,0x00,0x00);
    }
    else if (greens > blues)
    {
        colors[index] = GREEN;
        maxColor = CV_RGB(0x00,0xff,0x00);
    }
    else
    {
        maxColor = CV_RGB(0x00,0x00,0xff);
        colors[index] = PURPLE;
    }
    cvCircle(f_p, pt4, 10,  maxColor, 1);

    //cvCvtColor(frame_painted, gray, CV_BGR2GRAY);
    CBlobResult blobs;
    blobs = CBlobResult(mask, NULL, 100, true);
    // Now, must get rid of the fake null blob and background blob...
    blobs.Filter(blobs, B_INCLUDE, CBlobGetMean(), B_GREATER, 200);
    // Then, filter things down to the actual blobs...
    blobs.Filter(blobs, B_INCLUDE, CBlobGetArea(), B_GREATER, 100);
    int numBlobs = blobs.GetNumBlobs();
    numbers[index] = numBlobs;
    CBlob obj;
    CvPoint objPoint;
    objPoint.y = (pt4.y + pt5.y) / 2;
    for (int objNum = 0; objNum < numBlobs; objNum++)
    {
        blobs.GetNthBlob(CBlobGetArea(), objNum, obj);
        objPoint.x = int(CBlobGetXCenter()(obj));
        obj.FillBlob(f_p, CV_RGB(100, 100, 100));
        CvScalar shapeColor;
        if (obj.SumY() - (obj.MaxY() + obj.MinY())/2 > 3)
        {
            shapes[index] = DIAMOND;
            shapeColor = CV_RGB(0xff, 0x00, 0x00);
        }
        else if ((obj.MaxX() + obj.MinX())/2 - obj.SumX() > 2)
        {
            shapeColor = CV_RGB(0x00, 0xff, 0x00);
            shapes[index] = SQUIGGLE;
        }
        else
        {
            shapeColor = CV_RGB(0x00, 0x00, 0xff);
            shapes[index] = OVAL;
        }
        cvCircle(f_p, objPoint, 10, shapeColor, 2);
        CvPoint centersPoint;
        centersPoint.x = (obj.MaxX() + obj.MinX())/2;
        centersPoint.y = (obj.MaxY() + obj.MinY())/2;
        cvCircle(f_p, centersPoint, 2, CV_RGB(0x00, 0x90, 0xff), 1);
        centersPoint.x = int(obj.SumX());
        centersPoint.y = int(obj.SumY());
        cvCircle(f_p, centersPoint, 2, CV_RGB(0xff, 0x90, 0x00), 1);

        CvPoint shadingHi, shadingLo;
        int x = (obj.MaxX() + obj.MinX())/2;
        shadingHi.x = x+2;
        shadingLo.x = x-2;
        shadingHi.y = frame->height/2 + 5 + 1;
        shadingLo.y = frame->height/2 - 5 - 1;
        uchar* temp_ptr = &((uchar*)(frame->imageData +
                    frame->widthStep*(frame->height/2)))[x*3];
        int shadingMax = (temp_ptr[0] + temp_ptr[1] + temp_ptr[2]) / 3;
        int shadingMin = shadingMax;
        for (int y = frame->height/2 - 5; y <= frame->height/2 + 5; y++)
        {
            uchar* temp_ptr = &((uchar*)(frame->imageData +
                        frame->widthStep*y))[x*3];
            int val = (temp_ptr[0] + temp_ptr[1] + temp_ptr[2]) / 3;
            if (shadingMax < val) shadingMax = val;
            if (shadingMin > val) shadingMin = val;
        }
        int shading;
        if (shadingMax - shadingMin > 50)
        {
            shading = 2; // Mixed.
            textures[index] = STRIPED;
        }
        else if (shadingMax > 50)
        {
            shading = 1; // All light/empty.
            textures[index] = EMPTY;
        }
        else
        {
            shading = 3; // All dark/filled.
            textures[index] = FILLED;
        }
        cvRectangle(f_p, shadingLo, shadingHi, CV_RGB(250,0,250), shading);
    }   
    cvReleaseImage(&mask);
}

void setGlobals()
{
    // mouse click locations (old and new, to tell if we've
    // clicked on a new point
    old_click_pt.x = 0;
    old_click_pt.y = 0;
    new_click_pt.x = 0;
    new_click_pt.y = 0;

    // in theory these should get deleted at the end of the program, too...

    // these sizes are hard-coded, but should probably be obtained for each
    // image individually -- however, with our cameras and images, this is safe.

    // set up some global points for drawing, etc.
    // Points 4 and 5 are the box that we color in.
    pt4.x = 20;//12; 
    pt4.y = 35;//12;

    pt5.x = 300;//300;
    pt5.y = 90;//228;

}

/*
// this determines whether a pixel is Color or not...
bool isColor(int r, int g, int b, double hue, double sat, double val)
{
    return ((((hue >= Hue_Min) || (Hue_Min < 0 && hue >= Hue_Min + 1)) &&
                (hue <= Hue_Max))  || hue == -1) &&
        (sat >= Sat_Min) && (sat <= Sat_Max) &&
        (val >= Val_Min) && (val <= Val_Max);
}
*/
bool isRed(double hue, double sat, double val, double r, double g, double b)
{
    double Hue_Min = -0.1;
    double Hue_Max = 0.1;
    double Sat_Min = 0.3;
    double Sat_Max = 1.0;
    double Val_Min = 0.0;
    double Val_Max = 0.5;
    return ((((hue >= Hue_Min) || (Hue_Min < 0 && hue >= Hue_Min + 1)) &&
                (hue <= Hue_Max))  || hue == -1) &&
        (sat >= Sat_Min) && (sat <= Sat_Max) &&
        (val >= Val_Min) && (val <= Val_Max);
}

bool isGreen(double hue, double sat, double val, double r, double g, double b)
{
    double Hue_Min = 0.25;
    double Hue_Max = 0.45;
    double Sat_Min = 0.4;
    double Sat_Max = 1.0;
    double Val_Min = 0.0;
    double Val_Max = 0.5;
    return ((((hue >= Hue_Min) || (Hue_Min < 0 && hue >= Hue_Min + 1)) &&
                (hue <= Hue_Max))  || hue == -1) &&
        (sat >= Sat_Min) && (sat <= Sat_Max) &&
        (val >= Val_Min) && (val <= Val_Max);
}

bool isPurple(double hue, double sat, double val, double r, double g, double b)
{
    double Hue_Min = 0.45;
    double Hue_Max = 0.75;
    double Sat_Min = 0.3;
    double Sat_Max = 1.0;
    double Val_Min = 0.0;
    double Val_Max = 0.5;
    return ((((hue >= Hue_Min) || (Hue_Min < 0 && hue >= Hue_Min + 1)) &&
                (hue <= Hue_Max))  || hue == -1) &&
        (sat >= Sat_Min) && (sat <= Sat_Max) &&
        (val >= Val_Min) && (val <= Val_Max);
}
